#include "System.h"
#include "Registry.h"

CRegistry::CRegistry(HKEY root_key)
	: root_key(root_key), key(NULL)
{
}

CRegistry::~CRegistry()
{
	CloseKey();
}

#define CRegistry_CheckResult(result, message) if((result) != ERROR_SUCCESS) throw exception(message)

void CRegistry::OpenKey(const string &name, bool write, bool create)
{
	CloseKey();
	if(create)
	{
		CRegistry_CheckResult(RegCreateKeyEx(root_key, name, 0, NULL, REG_OPTION_NON_VOLATILE,
			write ? KEY_WRITE : KEY_READ, NULL, &key, NULL), "Can not create registry key");
	}
	else
	{
		CRegistry_CheckResult(RegOpenKeyEx(root_key, name, 0, write ? KEY_WRITE : KEY_READ, &key),
			"Can not open registry key");
	}
	key_name = name;
}

void CRegistry::CloseKey()
{
	if(key)
	{
		CRegistry_CheckResult(RegCloseKey(key), "Can not close registry key");
		key = NULL;
		key_name.Clear();
	}
}

void CRegistry::FlushKey()
{
	CRegistry_CheckResult(RegFlushKey(Key()), "Can not flush registry key");
}

void CRegistry::DeleteSubKey(const string &name)
{
	CRegistry_CheckResult(RegDeleteKey(Key(), name), "Can not delete registry key");
}

void CRegistry::GetSubKeyNames(strings &names)
{
	DWORD count;
	DWORD length;
	CRegistry_CheckResult(RegQueryInfoKey(Key(), NULL, NULL, NULL, &count, &length, NULL, NULL, NULL, NULL, NULL, NULL),
		"Can not retrieve registry key info");
	length++;
	TCHAR *name = new TCHAR[length];
	DWORD name_length;
	names.Clear();
	for(DWORD i = 0; i < count; i++)
	{
		name_length = length;
		LONG result = RegEnumKeyEx(Key(), i, name, &name_length, NULL, NULL, 0, NULL);
		if(result == ERROR_SUCCESS || result == ERROR_MORE_DATA) names.Add(string(name));
		else throw exception("Can not retrieve registry subkeys");
	}
}

int CRegistry::SubKeyCount()
{
	DWORD count;
	CRegistry_CheckResult(RegQueryInfoKey(Key(), NULL, NULL, NULL, &count, NULL, NULL, NULL, NULL, NULL, NULL, NULL),
		"Can not retrieve registry subkey count");
	return count;
}

string CRegistry::SubKeyName(int index)
{
	if(index < 0) throw exception("Invalid registry subkey index");
	DWORD count;
	DWORD length;
	CRegistry_CheckResult(RegQueryInfoKey(Key(), NULL, NULL, NULL, &count, &length, NULL, NULL, NULL, NULL, NULL, NULL),
		"Can not retrieve registry key info");
	if((DWORD)index >= count) throw exception("Invalid registry subkey index");
	string name(length);
	length++;
	LONG result = RegEnumKeyEx(Key(), index, name, &length, NULL, NULL, 0, NULL);
	if(result != ERROR_SUCCESS && result != ERROR_MORE_DATA) throw exception("Can not retrieve registry subkey name");
	name.Update();
	return name;
}

int CRegistry::SubKeyIndex(const string &name)
{
	for(int i = 0, c = SubKeyCount(); i < c; i++)
		if(name.Equals(SubKeyName(i), true)) return i;
	return -1;
}

bool CRegistry::SubKeyExists(const string &name)
{
	return SubKeyIndex(name) != -1;
}

void CRegistry::DeleteValue(const string &name)
{
	CRegistry_CheckResult(RegDeleteValue(Key(), name), "Can not delete registry value");
}

void CRegistry::GetValueNames(strings &names)
{
	DWORD count;
	DWORD length;
	CRegistry_CheckResult(RegQueryInfoKey(Key(), NULL, NULL, NULL, NULL, NULL, NULL, &count, &length, NULL, NULL, NULL),
		"Can not retrieve registry key info");
	length++;
	TCHAR *name = new TCHAR[length];
	DWORD name_length;
	names.Clear();
	for(DWORD i = 0; i < count; i++)
	{
		name_length = length;
		CRegistry_CheckResult(RegEnumValue(Key(), i, name, &name_length, NULL, NULL, NULL, NULL),
			"Can not retrieve registry values");
		names.Add(string(name));
	}
}

int CRegistry::ValueCount()
{
	DWORD count;
	CRegistry_CheckResult(RegQueryInfoKey(Key(), NULL, NULL, NULL, NULL, NULL, NULL, &count, NULL, NULL, NULL, NULL),
		"Can not retrieve registry value count");
	return count;
}

string CRegistry::ValueName(int index)
{
	if(index < 0) throw exception("Invalid registry value index");
	DWORD count;
	DWORD length;
	CRegistry_CheckResult(RegQueryInfoKey(Key(), NULL, NULL, NULL, NULL, NULL, NULL, &count, &length, NULL, NULL, NULL),
		"Can not retrieve registry key info");
	if((DWORD)index >= count) throw exception("Invalid registry value index");
	string name(length);
	length++;
	CRegistry_CheckResult(RegEnumValue(Key(), index, name, &length, NULL, NULL, NULL, NULL),
		"Can not retrieve registry value name");
	name.Update();
	return name;
}

DWORD CRegistry::ValueType(int index)
{
	CheckValueIndex(index);
	DWORD type;
	CRegistry_CheckResult(RegEnumValue(Key(), index, NULL, NULL, NULL, &type, NULL, NULL),
		"Can not retrieve registry value type");
	return type;
}

DWORD CRegistry::ValueType(const string &name)
{
	DWORD type;
	CRegistry_CheckResult(RegQueryValueEx(Key(), name, NULL, &type, NULL, NULL),
		"Can not retrieve registry value type");
	return type;
}

DWORD CRegistry::ValueSize(int index)
{
	CheckValueIndex(index);
	DWORD size;
	CRegistry_CheckResult(RegEnumValue(Key(), index, NULL, NULL, NULL, NULL, NULL, &size),
		"Can not retrieve registry value size");
	return size;
}

DWORD CRegistry::ValueSize(const string &name)
{
	DWORD size;
	CRegistry_CheckResult(RegQueryValueEx(Key(), name, NULL, NULL, NULL, &size),
		"Can not retrieve registry value size");
	return size;
}

void CRegistry::GetValue(int index, BYTE *data, DWORD &size, DWORD *type)
{
	CheckValueIndex(index);
	CRegistry_CheckResult(RegEnumValue(Key(), index, NULL, NULL, NULL, type, data, &size),
		"Can not retrieve registry value");
}

void CRegistry::GetValue(const string &name, BYTE *data, DWORD &size, DWORD *type)
{
	CRegistry_CheckResult(RegQueryValueEx(Key(), name, NULL, type, data, &size),
		"Can not retrieve registry value");
}

int CRegistry::ValueIndex(const string &name)
{
	for(int i = 0, c = ValueCount(); i < c; i++)
		if(name.Equals(ValueName(i), true)) return i;
	return -1;
}

bool CRegistry::ValueExists(const string &name)
{
	return ValueIndex(name) != -1;
}

DWORD CRegistry::GetDWordValue(const string &name)
{
	DWORD type;
	DWORD value;
	DWORD size = sizeof(DWORD);
	GetValue(name, (BYTE*)&value, size, &type);
	if(type != REG_DWORD) throw exception("Registry value type is not DWORD");
	if(size != sizeof(DWORD)) throw exception("Error retrieving registry value");
	return value;
}

QWORD CRegistry::GetQWordValue(const string &name)
{
	DWORD type;
	QWORD value;
	DWORD size = sizeof(QWORD);
	GetValue(name, (BYTE*)&value, size, &type);
	if(type != REG_QWORD) throw exception("Registry value type is not QWORD");
	if(size != sizeof(QWORD)) throw exception("Error retrieving registry value");
	return value;
}

int CRegistry::GetIntValue(const string &name)
{
	return (int)GetDWordValue(name);
}

bool CRegistry::GetBoolValue(const string &name)
{
	return GetDWordValue(name) != 0;
}

float CRegistry::GetFloatValue(const string &name)
{
	DWORD v = GetDWordValue(name);
	float value;
	memcpy(&value, &v, sizeof(float));
	return value;
}

double CRegistry::GetDoubleValue(const string &name)
{
	QWORD v = GetQWordValue(name);
	double value;
	memcpy(&value, &v, sizeof(double));
	return value;
}

string CRegistry::GetStringValue(const string &name)
{
	DWORD type;
	DWORD size;
	GetValue(name, NULL, size, &type);
	if(type != REG_EXPAND_SZ && type != REG_MULTI_SZ && type != REG_SZ)
		throw exception("Registry value type is not string");
	if(!size) throw exception("Registry value is zero size string");
	string value(size / sizeof(TCHAR) - 1);
	GetValue(name, (BYTE*)(LPTSTR)value, size, NULL);
	value.Update();
	return value;
}

DWORD CRegistry::GetDWordValue(const string &name, DWORD default_value)
{
	try
	{
		return GetDWordValue(name);
	}
	catch(exception &)
	{
		return default_value;
	}
}

QWORD CRegistry::GetQWordValue(const string &name, QWORD default_value)
{
	try
	{
		return GetQWordValue(name);
	}
	catch(exception &)
	{
		return default_value;
	}
}

int CRegistry::GetIntValue(const string &name, int default_value)
{
	try
	{
		return GetIntValue(name);
	}
	catch(exception &)
	{
		return default_value;
	}
}

bool CRegistry::GetBoolValue(const string &name, bool default_value)
{
	try
	{
		return GetBoolValue(name);
	}
	catch(exception &)
	{
		return default_value;
	}
}

float CRegistry::GetFloatValue(const string &name, float default_value)
{
	try
	{
		return GetFloatValue(name);
	}
	catch(exception &)
	{
		return default_value;
	}
}

double CRegistry::GetDoubleValue(const string &name, double default_value)
{
	try
	{
		return GetDoubleValue(name);
	}
	catch(exception &)
	{
		// Try binary read
		try
		{
			double value;
			DWORD type;
			DWORD size = sizeof(double);
			GetValue(name, (BYTE*)&value, size, &type);
			if(type != REG_BINARY || size != sizeof(double)) throw exception();
			return value;
		}
		catch(exception &)
		{
			return default_value;
		}
	}
}

string CRegistry::GetStringValue(const string &name, const string &default_value)
{
	try
	{
		return GetStringValue(name);
	}
	catch(exception &)
	{
		return default_value;
	}
}

void CRegistry::SetValue(const string &name, const BYTE *data, DWORD size, DWORD type)
{
	CRegistry_CheckResult(RegSetValueEx(Key(), name, 0, type, data, size),
		"Can not set registry value");
}

void CRegistry::SetDWordValue(const string &name, DWORD value)
{
	SetValue(name, (BYTE*)&value, sizeof(DWORD), REG_DWORD);
}

void CRegistry::SetQWordValue(const string &name, QWORD value)
{
	SetValue(name, (BYTE*)&value, sizeof(QWORD), REG_QWORD);
}

void CRegistry::SetIntValue(const string &name, int value)
{
	SetDWordValue(name, (DWORD)value);
}

void CRegistry::SetBoolValue(const string &name, bool value)
{
	SetDWordValue(name, (DWORD)value);
}

void CRegistry::SetFloatValue(const string &name, float value)
{
	DWORD v;
	memcpy(&v, &value, sizeof(float));
	SetDWordValue(name, v);
}

void CRegistry::SetDoubleValue(const string &name, double value)
{
	QWORD v;
	memcpy(&v, &value, sizeof(double));
	SetQWordValue(name, v);
}

void CRegistry::SetStringValue(const string &name, const string &value)
{
	SetValue(name, (const BYTE*)(LPCTSTR)value, (value.Length() + 1) * sizeof(TCHAR), REG_SZ);
}
