#pragma once

// Arithmetic routines
#ifndef PI
const double PI = 3.1415926535897932384626433832795;
#endif
inline int Round(double value) { return (int)(value < 0 ? value - 0.5 : value + 0.5); }
inline int RoundP(double value) { return (int)(value + 0.5); }
inline double PercentF(int count, int total) { return 100.0 * count / total; }
inline double PercentFZ(int count, int total) { return total ? 100.0 * count / total : 0.0; }
inline int Percent(int count, int total) { return Round(PercentF(count, total)); }
inline int PercentZ(int count, int total) { return Round(PercentFZ(count, total)); }

// Flag routines
inline bool IsFlagSet(DWORD flags, DWORD flag)
{
	return (flags & flag) == flag;
}

inline void SetFlag(DWORD &flags, DWORD flag)
{
	flags |= flag;
}

inline void ClearFlag(DWORD &flags, DWORD flag)
{
	flags &= ~flag;
}

inline void SetFlag(DWORD &flags, DWORD flag, bool value)
{
	if(value) SetFlag(flags, flag); else ClearFlag(flags, flag);
}

// String routines
string Format(LPCTSTR format, ...);
string Format(DWORD id, ...);
string FormatV(LPCTSTR format, va_list list);
string FormatV(DWORD id, va_list list);
string AdjustLineBreaks(const string &s);
string GetOneLine(const string &s);

// Exceptions
inline exception EXCEPTION(DWORD id, ...)
{
	va_list list;
	va_start(list, id);
	exception e = EXCEPTION(Format(id, list));
	va_end(list);
	return e;
}

inline string FormatComError(_com_error &e)
{
	return Format(TEXT("COM error:\n\tCode = %08lx\n\tCode meaning = %s\n\tSource = %s\n\tDescription = %s"),
		e.Error(), e.ErrorMessage(), (LPCTSTR)e.Source(), (LPCTSTR)e.Description());
}

inline string FormatComError(_com_error &e, const string &where)
{
	return Format(TEXT("%s\n")
		TEXT("COM error:\n\tCode = %08lx\n\tCode meaning = %s\n\tSource = %s\n\tDescription = %s"),
		(LPCTSTR)where,
		e.Error(), e.ErrorMessage(), (LPCTSTR)e.Source(), (LPCTSTR)e.Description());
}

inline void ComErrorToException(_com_error &e)
{
	throw EXCEPTION(FormatComError(e));
}

inline void ComErrorToException(_com_error &e, const string &where)
{
	throw EXCEPTION(FormatComError(e, where));
}

// File name and path routines
extern const TCHAR directorySeparator;
extern const TCHAR altDirectorySeparator;
extern const TCHAR volumeSeparator;

bool IsValidPathChar(TCHAR c);
void CheckValidPathChar(TCHAR c);
void CheckValidPathToken(const string &s);

string GetCleanedPath(const string &path);
string GetCurrentDirectory();
string GetFileExtension(const string &filename);
string GetFilePathWithoutExtension(const string &filename);
bool FileNameHasNoExtension(const string &filename);
string ChangeFileExtension(const string &filename, const string &ext);
string GetFileName(const string &filename);
string GetFileNameWithoutExtension(const string &filename);
bool IsUNCPath(const string &path);
string GetDriveName(const string &filename);
string GetDirectoryName(const string &filename);
string GetSubdirectories(const string &path);
void GetSubdirectories(const string &filename, strings &dirs);
string GetSubdirectoryName(const string &filename, int index);
string MakeFullFileName(const string &path, const string &filename);
string MakeRelativeFileName(const string &path, const string &filename);
string GetFullFileName(const string &filename);
string GetRelativeFileName(const string &filename);
bool FileExists(const string &filename);
void AddTrailingBackslash(string &path);
void RemoveTrailingBackslash(string &path);
string FindFile(const string &path, const string &filter = STRING("*.*"), bool search_subdirectories = true);

// File filter routines
string MakeFileFilterString(const string &name, const string &filter, bool filter_in_name = true);
string AllFilesFilter();
string AllFilesFilterName();
string GetAllFilesFilterString(bool filter_in_name = true);
bool FileFilterHasExtension(const string &filter, const string &ext);
int ExtensionsInFileFilter(const string &filter);
string GetExtensionFromFileFilter(const string &filter, int index = 0);
void AddFileFilterString(string &s, const string &filter_string);
void AddFileFilterString(string &s, const string &name, const string &filter, bool filter_in_name = true);
void AddAllFilesFilterString(string &s, bool filter_in_name = true);

// Convertion to string routines

extern const string falseString;
extern const string trueString;

string DateToString();
string TimeToString();
string IntToString(int value, int base = 10);
string IntToString(int value, const string &one, const string &many = emptyString, const string &no = emptyString);
string HexToString(int value);
string OctToString(int value);
string BinToString(int value);
string BoolToString(bool value, const string &trueValue = trueString, const string &falseValue = falseString);
string FloatToString(double value);
string FloatToString(double value, int digits);
string FloatToStringS(double value, int digits);

struct MateValue
{
private:
	bool many;
	string name;
	string names;
	int maxValue;
	int value;

	void SetName(const string &value)
	{
		if(value.IsEmpty()) throw exception("Mate name can not be an empty string");
		name = value;
	}

	void SetMaxValue(int value)
	{
		if(value < 0) value = -1;
		if(value != -1 && value <= 1) throw exception("Mate max value have to be greater than 1");
		this->maxValue = value;
	}

public:
	MateValue(const string &name = STRING("mate"), int maxValue = 1000, int value = 0)
	{
		many = false;
		SetName(name);
		SetMaxValue(maxValue);
		SetValue(value);
	}

	MateValue(const string &name, const string &names, int maxValue = 1000, int value = 0)
	{
		many = true;
		SetName(name);
		this->names = names;
		SetMaxValue(maxValue);
		SetValue(value);
	}

	string ToString() const
	{
		return many ? IntToString(value, name, names) : IntToString(value) + TEXT(' ') + name;
	}

	const string & Name() const { return name; }
	int MaxValue() const { return maxValue; }

	int Value() const { return value; }
	void SetValue(int value)
	{
		if(value < 0) throw exception("Mate value can not be less than zero");
		if(maxValue != -1 && value >= maxValue) throw exception("Mate value have to be less than max value");
		this->value = value;
	}
};

string MateValuesToString(const vector<MateValue> &mateValues);
void MinimizeMateValues(const vector<MateValue> &mateValues, vector<MateValue> &dstMateValues, int maxMates);
string MateValuesToString(const vector<MateValue> &mateValues, int maxMates);
void SetMateValues(DWORD value, vector<MateValue> &mateValues);
void SetMateValues(unsigned __int64 value, vector<MateValue> &mateValues);
string MatedValueToString(DWORD value, vector<MateValue> &mateValues, int maxMates = 2);
string MatedValueToString(unsigned __int64 value, vector<MateValue> &mateValues, int maxMates = 2);

string TimeToString(DWORD ms, int maxMates = 2);
string TimeToString(double sec, int maxMates = 2);
string SizeToString(DWORD bytes, int maxMates = 2);

// Convertion from string routines
bool StringToBool(const string &value, const string &trueValue = trueString, const string &falseValue = falseString);
int StringToInt(const string &value, const string &end = emptyString, int base = 10);
int StringToHex(const string &value, const string &end = emptyString);
int StringToOct(const string &value, const string &end = emptyString);
int StringToBin(const string &value, const string &end = emptyString);
double StringToFloat(const string &value, const string &end = emptyString);

// High resolution timer routines
inline double PerformanceCounterToSeconds(__int64 tc)
{
	static bool initialized = false;
	static __int64 f;
	if(!initialized)
	{
		if(!QueryPerformanceFrequency((LARGE_INTEGER*)&f)) throw exception("Can not query performance frequency");
		initialized = true;
	}
	return (double)tc / (double)f;
}

inline double GetTickCountEx()
{
	__int64 tc;
	if(!QueryPerformanceCounter((LARGE_INTEGER*)&tc)) throw exception("Can not query performance counter");
	return PerformanceCounterToSeconds(tc);
}
