#include "System.h"
#include "VFFileList.h"
#include "Utils.h"

const string defaultFileListFileExtension = TEXT("vffl");

string GetVFFileListFilter()
{
	return TEXT("*.") + defaultFileListFileExtension;
}

string GetVFFileListFilterName()
{
	return TEXT("File lists");
}

string GetVFFileListFilterString(bool filter_in_name)
{
	return MakeFileFilterString(GetVFFileListFilterName(), GetVFFileListFilter(), filter_in_name);
}

string GetLoadVFFileListFilterString(bool filter_in_name, bool all_files)
{
	string s;
	AddFileFilterString(s, GetVFFileListFilterString(filter_in_name));
	if(all_files) AddAllFilesFilterString(s, filter_in_name);
	return s;
}

string GetSaveVFFileListFilterString(bool filter_in_name)
{
	string s;
	AddFileFilterString(s, GetVFFileListFilterString(filter_in_name));
	return s;
}

bool LoadVFFileList(const string &filename, VFFileList *filelist)
{
	FILE *f = _tfopen(filename, TEXT("rt"));
	if(!f) return false;
	bool result = false;
	filelist->Clear();
	const int max_line_length = 1000;
	TCHAR buf[max_line_length + 1];
	while(_fgetts(buf, max_line_length, f))
	{
		string s(buf);
		int i = s.IndexOf(TEXT('\n'));
		if(i != -1) s.SetLength(i);
		if(!s.IsEmpty())
		{
			string filename;
			int resolution = VF_IMAGE_RESOLUTION;
			string finger_id;
			int i = s.IndexOf(TEXT('\t'));
			if(i != -1)
			{
				int j = s.IndexOf(TEXT('\t'), i + 1);
				if(j != -1)
				{
					finger_id = s.Substring(j + 1);
				}
				else j = s.Length();
				try
				{
					resolution = StringToInt(s.Substring(i + 1, j - i - 1));
				}
				catch(exception &)
				{
				}
			}
			else i = s.Length();
			filename = s.Substring(0, i);
			if(!filename.IsEmpty()) filelist->Add(VFFileListItem(filename, resolution, finger_id));
		}
	}
	result = true;
	fclose(f);
	return result;
}

bool SaveVFFileList(const string &filename, VFFileList *filelist)
{
	FILE *f = _tfopen(filename, TEXT("wt"));
	if(!f) return false;
	for(VFFileListItem* i = filelist->Begin(); i < filelist->End(); i++)
		if(_fputts(Format(TEXT("%s\t%d\t%s"), (LPCTSTR)i->filename, i->resolution, (LPCTSTR)i->finger_id), f) < 0
			|| (i != filelist->End() - 1 && _fputts(TEXT("\n"), f) < 0))
		{
			fclose(f);
			return false;
		}
	fclose(f);
	return true;
}

string GetVFFileListRoot(VFFileList &fileList)
{
	if(fileList.IsEmpty()) return emptyString;
	string s = fileList.First().filename;
	bool rootUNC = IsUNCPath(s);
	string rootDrive = GetDriveName(s);
	strings root;
	GetSubdirectories(s, root);
	strings ss;
	for(VFFileListItem * i = fileList.Begin() + 1; i < fileList.End() && !root.IsEmpty(); i++)
	{
		s = i->filename;
		if(IsUNCPath(s) != rootUNC || GetDriveName(s) != rootDrive) root.Clear();
		else
		{
			GetSubdirectories(s, ss);
			int j;
			for(j = 0; j < root.Length() && j < ss.Length(); j++)
			{
				if(!root[j].Equals(ss[j], true)) break;
			}
			root.SetLength(j);
		}
	}
	s.Clear();
	if(!rootDrive.IsEmpty())
	{
		s += rootDrive;
		s += rootUNC ? string(2, directorySeparator) : string(1, volumeSeparator) + directorySeparator;
	}
	for(string * pS = root.Begin(); pS < root.End(); pS++)
	{
		s += *pS;
		s += directorySeparator;
	}
	return s;
}

/////////////////////////
//                     //
// Finger ID functions //
//                     //
/////////////////////////

const string default_file_finger_id(TEXT("$(NameNoExt)"));
const string default_scanner_finger_id(TEXT("$(Name) $(Date) $(Time)"));

string FingerID_GetArgument(const string &arguments, int i);
string FingerID_TranslateFunction(const string &expression, const string &source, bool from_scanner);

string FingerID_Path(bool /*from_scanner*/, const string & /*arguments*/, const string &filename)
{
	return filename;
}

string FingerID_Drive(bool /*from_scanner*/, const string & /*arguments*/, const string &filename)
{
	return GetDriveName(filename);
}

string FingerID_Dir(bool /*from_scanner*/, const string & /*arguments*/, const string &filename)
{
	return GetDirectoryName(filename);
}

string FingerID_Name(bool /*from_scanner*/, const string & /*arguments*/, const string &filename)
{
	return GetFileName(filename);
}

string FingerID_Ext(bool /*from_scanner*/, const string & /*arguments*/, const string &filename)
{
	return GetFileExtension(filename);
}

string FingerID_NameNoExt(bool /*from_scanner*/, const string & /*arguments*/, const string &filename)
{
	return GetFileNameWithoutExtension(filename);
}

string FingerID_PathNoExt(bool /*from_scanner*/, const string & /*arguments*/, const string &filename)
{
	return GetFilePathWithoutExtension(filename);
}

string FingerID_Subdir(bool /*from_scanner*/, const string &arguments, const string &filename)
{
	int i = StringToInt(FingerID_GetArgument(arguments, 0));
	return GetSubdirectoryName(filename, i);
}

string FingerID_ScannerFullName(bool /*from_scanner*/, const string & /*arguments*/, const string &scanner_id)
{
	return scanner_id;
}

string FingerID_ScannerName(bool /*from_scanner*/, const string & /*arguments*/, const string &scanner_id)
{
	int i = scanner_id.IndexOf(TEXT('\\'));
	if(i == -1) i = scanner_id.Length();
	return scanner_id.Substring(0, i);
}

string FingerID_ScannerID(bool /*from_scanner*/, const string & /*arguments*/, const string &scanner_id)
{
	int i = scanner_id.IndexOf(TEXT('\\'));
	return i == -1 ? emptyString : scanner_id.Substring(i + 1);
}

string FingerID_Left(bool from_scanner, const string &arguments, const string &str)
{
	string s = FingerID_TranslateFunction(FingerID_GetArgument(arguments, 0), str, from_scanner);
	int c = StringToInt(FingerID_GetArgument(arguments, 1));
	if(abs(c) > s.Length()) c = c < 0 ? -s.Length() : s.Length();
	return c >= 0 ? s.Substring(0, c) : s.Substring(-c);
}

string FingerID_Right(bool from_scanner, const string &arguments, const string &str)
{
	string s = FingerID_TranslateFunction(FingerID_GetArgument(arguments, 0), str, from_scanner);
	int c = StringToInt(FingerID_GetArgument(arguments, 1));
	if(abs(c) > s.Length()) c = c < 0 ? -s.Length() : s.Length();
	return c >= 0 ? s.Substring(s.Length() - c) : s.Substring(0, s.Length() + c);
}

string FingerID_Center(bool from_scanner, const string &arguments, const string &str)
{
	string s = FingerID_TranslateFunction(FingerID_GetArgument(arguments, 0), str, from_scanner);
	int i = StringToInt(FingerID_GetArgument(arguments, 1));
	if(i == 0) return s;
	if(i < 0)
	{
		i = s.Length() + i;
		if(i < 0) i = 0;
	}
	else
	{
		i--;
		if(i > s.Length()) i = s.Length();
	}
	int c = StringToInt(FingerID_GetArgument(arguments, 2));
	return c >= 0 ? s.Substring(i, c) : s.Substring(0, i) + s.Substring(min(i - c, s.Length()));
}

string FingerID_Number(bool from_scanner, const string &arguments, const string &str)
{
	string s = FingerID_TranslateFunction(FingerID_GetArgument(arguments, 0), str, from_scanner);
	int c = StringToInt(FingerID_GetArgument(arguments, 1));
	if(c == 0) return emptyString;
	else if(c > 0)
	{
		for(int i = 0, j = -1; i <= s.Length() + 1; i++)
		{
			if(i < s.Length() && IsDigit(s[i]))
			{
				if(j == -1) j = i;
			}
			else if(j != -1)
			{
				if(!--c) return s.Substring(j, i - j);
				j = -1;
			}
		}
		return emptyString;
	}
	else
	{
		for(int i = s.Length() - 1, j = -1; i >= -1; i--)
		{
			if(i >= 0 && IsDigit(s[i]))
			{
				if(j == -1) j = i;
			}
			else if(j != -1)
			{
				if(!++c) return s.Substring(i + 1, j - i);
				j = -1;
			}
		}
		return emptyString;
	}
}

string FingerID_NoNumber(bool from_scanner, const string &arguments, const string &str)
{
	string s = FingerID_TranslateFunction(FingerID_GetArgument(arguments, 0), str, from_scanner);
	int c = StringToInt(FingerID_GetArgument(arguments, 1));
	if(c == 0) return s;
	else if(c > 0)
	{
		for(int i = 0, j = -1; i <= s.Length() + 1; i++)
		{
			if(i < s.Length() && IsDigit(s[i]))
			{
				if(j == -1) j = i;
			}
			else if(j != -1)
			{
				if(!--c) return s.LeftSubstring(j) + s.Substring(i);
				j = -1;
			}
		}
		return s;
	}
	else
	{
		for(int i = s.Length() - 1, j = -1; i >= -1; i--)
		{
			if(i >= 0 && IsDigit(s[i]))
			{
				if(j == -1) j = i;
			}
			else if(j != -1)
			{
				if(!++c) return s.LeftSubstring(i + 1) + s.Substring(j + 1);
				j = -1;
			}
		}
		return s;
	}
}

string FingerID_Date(bool /*from_scanner*/, const string & /*arguments*/, const string &/*str*/)
{
	return DateToString();
}

string FingerID_Time(bool /*from_scanner*/, const string & /*arguments*/, const string &/*str*/)
{
	return TimeToString();
}

struct FingerIDFunction
{
	const TCHAR *name;
	int argument_count;
	bool from_file;
	bool from_scanner;
	string (*proc)(bool from_scanner, const string &arguments, const string &filename);
	const TCHAR *help;
};

const FingerIDFunction finger_id_functions[] = {
	{ TEXT("Path"), 0, true, false, FingerID_Path, TEXT("Path - returns path to the file") },
	{ TEXT("Drive"), 0, true, false, FingerID_Drive, TEXT("Drive - returns drive letter from the path") },
	{ TEXT("Dir"), 0, true, false, FingerID_Dir, TEXT("Dir - returns file directory") },
	{ TEXT("Name"), 0, true, false, FingerID_Name, TEXT("Name - returns file name") },
	{ TEXT("Ext"), 0, true, false, FingerID_Ext, TEXT("Ext - returns file extension") },
	{ TEXT("NameNoExt"), 0, true, false, FingerID_NameNoExt, TEXT("NameNoExt - returns file name without extension") },
	{ TEXT("PathNoExt"), 0, true, false, FingerID_PathNoExt, TEXT("PathNoExt - returns file path without extension") },
	{ TEXT("Subdir"), 1, true, false, FingerID_Subdir, TEXT("SubDir(i) - if i = 0 returns all directories that the path contains\n")
		TEXT("\tif i > 0 returns i directory from the left of the path\n")
		TEXT("\tif i < 0 returns -i directory from the right of the path") },
	{ TEXT("FullName"), 0, false, true, FingerID_ScannerFullName, TEXT("FullName - returns full scanner name (with id)") },
	{ TEXT("Name"), 0, false, true, FingerID_ScannerName, TEXT("Name - returns scanner name") },
	{ TEXT("ID"), 0, false, true, FingerID_ScannerID, TEXT("ID - returns scanner id") },
	{ TEXT("Left"), 2, true, true, FingerID_Left, TEXT("Left(s, n) - if n > 0, returns at most n characters from the left of s\n")
		TEXT("\tif n < 0, returns s without at most -n characters from the left\n")
		TEXT("\tif n = 0, returns empty string") },
	{ TEXT("Right"), 2, true, true, FingerID_Right, TEXT("Right(s, n) - if n > 0, returns at most n characters from the right of s\n")
		TEXT("\tif n < 0, returns s without at most -n characters from the right\n")
		TEXT("\tif n = 0, returns empty string") },
	{ TEXT("Center"), 3, true, true, FingerID_Center, TEXT("Center(s, i, n) - if n > 0, returns at most n characters from the position\n")
		TEXT("\tif n < 0, returns s without at most -n characters from the position\n")
		TEXT("\tif n = 0, returns empty string\n")
		TEXT("\tposition is i from the left if i > 0 and -i from the right if i < 0\n")
		TEXT("\tif i = 0 returns s") },
	{ TEXT("Number"), 2, true, true, FingerID_Number, TEXT("Number(s, i) - if i = 0 returns empty string\n")
		TEXT("\tif i > 0, returns i-th number from the left of the string\n")
		TEXT("\tif i < 0, returns (-i)-th number from the right of the string") },
	{ TEXT("NoNumber"), 2, true, true, FingerID_NoNumber, TEXT("NoNumber(s, i) - if i = 0 returns s\n")
		TEXT("\tif i > 0, returns s without i-th number from the left\n")
		TEXT("\tif i < 0, returns s without (-i)-th number from the right") },
	{ TEXT("Date"), 0, false, true, FingerID_Date, TEXT("Date - returns current date") },
	{ TEXT("Time"), 0, false, true, FingerID_Time, TEXT("Time - returns current time") },
	{ NULL, 1, true, true, NULL, NULL }
};

string FingerID_GetArguments(const string &expression, const TCHAR * &i, const string &name, bool spaces_ok = true)
{
	try
	{
		const TCHAR *s = NULL, *e = NULL;
		int c = 0;
		for(const TCHAR *j = i; j < expression.End(); j++)
			if(*j == TEXT('('))
			{
				if(!c) s = j;
				c++;
			}
			else if(*j == TEXT(')'))
			{
				c--;
				if(c < 0) break;
				if(!c)
				{
					e = j;
					break;
				}
			}
			else if(c == 0)
				if(*j == TEXT(' '))
				{
					if(!spaces_ok) throw exception("Unexpected space");
				}
				else throw EXCEPTION(Format(TEXT("Unexpected char '%c'"), *j));
		if(c > 0) throw exception("Opening bracket without closing bracket");
		if(c < 0) throw exception("Closing bracket without opening bracket");
		if(!s || !e) throw exception("Brackets not found");
		i = e + 1;
		string ss = expression.Substring(s - expression.Begin() + 1, e - s - 1);
		ss.Trim();
		while(ss.Length() > 1 && ss[0] == TEXT('(') && ss[ss.Length() - 1] == TEXT(')'))
		{
			ss.Remove(ss.Length() - 1);
			ss.Remove(0);
			ss.Trim();
		}
		return ss;
	}
	catch(exception &e)
	{
		exception ee = EXCEPTION(Format(TEXT("%s after %s"), (LPCTSTR)AnsiToString(e.what()),
			name.IsEmpty() ? TEXT("macro") : (LPCTSTR)(TEXT("function ") + name)));
		throw ee;
	}
}

int FingerID_ArgumentCount(const string &arguments)
{
	if(arguments.IsEmpty()) return 0;
	int bc = 0, c = 0;
	for(const TCHAR *j = arguments.Begin(); j <= arguments.End(); j++)
	{
		if((j == arguments.End() || *j == TEXT(',')) && !bc) c++;
		else if(*j == TEXT('(')) bc++;
		else if(*j == TEXT(')')) bc--;
	}
	return c;
}

string FingerID_GetArgument(const string &arguments, int i)
{
	int c = 0, ii = i;
	for(const TCHAR *j = arguments.Begin(), *s = j; j <= arguments.End(); j++)
	{
		if((j == arguments.End() || *j == TEXT(',')) && !c)
		{
			if(!i)
			{
				string ss = arguments.Substring(s - arguments.Begin(), j - s);
				ss.Trim();
				while(ss.Length() > 1 && ss[0] == TEXT('(') && ss[ss.Length() - 1] == TEXT(')'))
				{
					ss.Remove(ss.Length() - 1);
					ss.Remove(0);
					ss.Trim();
				}
				return ss;
			}
			else
			{
				i--;
				s = j + 1;
			}
		}
		else if(*j == TEXT('(')) c++;
		else if(*j == TEXT(')')) c--;
	}
	throw EXCEPTION(Format(TEXT("Argument %d not found"), ii));
}

void FingerID_TestArgumentCount(const string &arguments, int c, const string &name)
{
	if(FingerID_ArgumentCount(arguments) != c)
	{
		exception ee = EXCEPTION(Format(TEXT("%s requires %s"),
			name.IsEmpty() ? TEXT("Macro") : (LPCTSTR)(name + TEXT(" function")),
			(LPCTSTR)IntToString(c, STRING("argument"), STRING("arguments"), STRING("no"))));
		throw ee;
	}
}

void FingerID_TestNoArguments(const string &expression, const TCHAR *i, const string &name)
{
	for(; i < expression.End(); i++)
		if(*i != TEXT(' ')) throw EXCEPTION(Format(TEXT("Unexpected char '%c' after function %s"), *i, (LPCTSTR)name));
}

string FingerID_TranslateFunction(const string &expression, const string &source, bool from_scanner)
{
	const TCHAR *e = expression.End();
	for(const TCHAR *i = expression.Begin(); i < expression.End(); i++)
		if(*i == TEXT('(') || *i == TEXT(')') || *i == TEXT(' '))
		{
			e = i;
			break;
		}
	string s = expression.Substring(0, e - expression.Begin());
	if(s.IsEmpty()) throw exception("Function name expected");
	const FingerIDFunction *f = finger_id_functions;
	string arguments;
	while(f->name)
	{
		if((from_scanner && f->from_scanner) || (!from_scanner && f->from_file))
			if(s.Equals(f->name, true))
			{
				if(f->argument_count)
				{
					arguments = FingerID_GetArguments(expression, e, f->name);
					FingerID_TestArgumentCount(arguments, f->argument_count, f->name);
				}
				else
				{
					FingerID_TestNoArguments(expression, e, f->name);
					arguments.Clear();
				}
				return f->proc(from_scanner, arguments, source);
			}
		f++;
	}
	throw EXCEPTION(Format(TEXT("Unknown function \"%s\""), (LPCTSTR)s));
}

const TCHAR finger_id_macro_char = TEXT('$');

string MakeFingerID(const string &finger_id, const string &source, bool from_scanner)
{
	string s;
	for(const TCHAR *i = finger_id.Begin(); i < finger_id.End(); i++)
		if(*i == finger_id_macro_char)
		{
			i++;
			if(i != finger_id.End() && *i == finger_id_macro_char)
			{
				s += finger_id_macro_char;
			}
			else
			{
				string arguments = FingerID_GetArguments(finger_id, i, emptyString, false);
				i--;
				FingerID_TestArgumentCount(arguments, 1, emptyString);
				s += FingerID_TranslateFunction(arguments, source, from_scanner);
			}
		}
		else if(*i == TEXT('/'))
		{
			if(i != finger_id.End() - 1 && *(i + 1) == TEXT('/'))
			{
				i++;
				s.Append(TEXT('/'));
			}
			else if(s.IsEmpty() || s[s.Length() - 1] != TEXT('\\')) s.Append(TEXT('\\'));
		}
		else s.Append(*i);
	return s;
}

void CheckFingerID(const string &finger_id, bool from_scanner)
{
	MakeFingerID(finger_id, emptyString, from_scanner);
}

string GetFingerIDHelp(bool from_scanner)
{
	const TCHAR *intro =
		TEXT("When enrolling fingerprints, finger id is stored in the database along with features extracted from the fingerprint image.\n")
		TEXT("Fingerprints from different fingers must have different finger id. Same finger fingerprints must have the same finger id.\n")
		TEXT("When identifying a fingerprint in the database you will get identified finger id as a result.\n")
		TEXT("You can specify finger id directly for each fingerprint, like \"John's thumb\" or use macros to make finger id from fingerprint image file name.\n")
		TEXT("Finger ID can contain any combination of characters and macros.\n")
		TEXT("Macros have to be in form of \"$(s)\", \"s\" can be one of the following functions:");
	const TCHAR *example =
		TEXT("Macros are especially usefull when enrolling a large amount of fingerprints from files\n")
		TEXT("For example, if you have fingerprint images in files that are named \"xxxxy.bmp\", where \"xxxx\" is finger number and \"y\" is scanning number,\n")
		TEXT("then you can specify \"$(Right(NameNoExt,-1))\" macro to get \"xxxx\" as finger id from the file name\n");
	/*const TCHAR *details =	
		TEXT("If you want to have '$' character in finger id, use \"$$\" instead.\n")
		TEXT("If you specify '/' character, then '\\' character is added to finger id if none is currently present.\n")
		TEXT("If you want to have the '/' character in finger id, use \"//\" instead.\n");*/

	const FingerIDFunction *f = finger_id_functions;
	string s(intro);
	s += TEXT('\n');
	while(f->name)
	{
		if((from_scanner && f->from_scanner) || (!from_scanner && f->from_file))
		{
			s += TEXT('\n');
			s += f->help;
		}
		f++;
	}
	s += TEXT("\n\ni and n parameters of the functions have to be an integer values.");
	if(!from_scanner)
	{
		s += TEXT("\n\n");
		s += example;
	}
	return s;
}
