import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.FilteredImageSource;
import java.awt.image.ImageProducer;

import neurotechnologija.utils.FeaturesFilter;
import neurotechnologija.utils.ImagePanel;

import VeriFinger.VeriFingerFeatures;
import VeriFinger.VeriFingerWrapper;

/**
 * ImagePanel extension for drawing image features over
 * generalized image
 *
 * @version 1.0 02/07/2002
 */
public class FeaturesImagePanel extends ImagePanel {
        private VeriFingerFeatures m_features = null;

        public FeaturesImagePanel() {
                super(null, 1.0);
        }

        public FeaturesImagePanel(Image img) {
                super(img, 1.0);
        }

        /**
         * Sets fingerprint features
         * */
        public void setData(VeriFingerFeatures data) {
                this.m_features = data;
        }

        /**
         * Overriden to draw generalized fingerprint image
         * */
        protected void drawImage(Graphics g) {
                Insets insets = this.getInsets();
                Rectangle bounds = this.getBounds();

                g.setColor(imageBackground);
                g.fillRect(
                        insets.left,
                        insets.top,
                        bounds.width - insets.left - insets.right,
                        bounds.height - insets.top - insets.bottom);

                if (image != null) {
                                g.setClip(
                                        insets.left + GAP,
                                        insets.top + GAP,
                                        bounds.width - insets.left - insets.right - GAP * 2,
                                        bounds.height - insets.top - insets.bottom - GAP * 2);
                                FeaturesFilter filter = new FeaturesFilter();
                                ImageProducer producer = new FilteredImageSource(image.getSource(), filter);
                                Image filteredImage = Toolkit.getDefaultToolkit().createImage(producer);
                                g.drawImage(
                                        filteredImage,
                                        insets.left + GAP,
                                        insets.top + GAP,
                                        imageWidth,
                                        imageHeight,
                                        null);
                                drawFeatures(g);
                }
        }

        /**
         * Draws fingerprint features
         * @param g
         */
        protected void drawFeatures(Graphics g) {
                if (m_features == null) return;

                int i, j, xx, yy;
                double x, y;
                Line2D.Double line;
                Ellipse2D.Double circle;
                Rectangle2D.Double rect;
                final double r = Math.PI / 180.0; // used in angles to radians conversion
                Insets insets = this.getInsets();
                Graphics2D g2 = (Graphics2D) g;

                AffineTransform at = new AffineTransform();
                at.translate(insets.left + GAP, insets.top + GAP);
                at.scale(zoom / 100.0, zoom / 100.0);
                g2.setColor(Color.red);
                g2.setStroke(new BasicStroke(2.0f));
                g2.transform(at);

                // draw fingerprint features
                for (i = 0; i < m_features.m_count; ++i) {
                        // convert to unsigned
                        int dir = (int) (m_features.m_d[i] & 0xFF);
                        double d = VeriFingerWrapper.VFDirToDeg(dir);
                        x = 10 * Math.cos(d * r);
                        y = 10 * Math.sin(d * r);
                        // direction
                        line = new Line2D.Double(
                                m_features.m_x[i],
                                m_features.m_y[i],
                                m_features.m_x[i] + x,
                                m_features.m_y[i] + y);
                        g2.draw(line);
                        // location
                        circle = new Ellipse2D.Double(
                                m_features.m_x[i] - 4,
                                m_features.m_y[i] - 4,
                                8, 8);
                        g2.draw(circle);
                }

                // draw fingerprint class data
                for(i = 0; i < m_features.sp_count; ++i) {
                                switch((int)m_features.sp_t[i]) {

                                case VeriFingerWrapper.VFFEAT_SP_CORE:
                                        // circle
                                        circle = new Ellipse2D.Double(
                                        m_features.sp_x[i] - 8,
                                        m_features.sp_y[i] - 8,
                                        16,     16);
                                        g2.draw(circle);
                                        break;

                                case VeriFingerWrapper.VFFEAT_SP_DOUBLECORE:
                                        // rectangle
                                        x = m_features.sp_x[i] - 8;
                                        y = m_features.sp_y[i] - 8;
                                        rect = new Rectangle2D.Double(x, y, 15, 15);
                                        g2.draw(rect);
                                        x = -16 * Math.cos(r * m_features.sp_d[i]);
                                        y = 16 * Math.sin(r * m_features.sp_d[i]);
                                        line = new Line2D.Double(
                                                m_features.sp_x[i],
                                                m_features.sp_y[i],
                                                m_features.sp_x[i] + x,
                                                m_features.sp_y[i] + y);
                                        g2.draw(line);
                                        break;

                                case VeriFingerWrapper.VFFEAT_SP_DELTA:
                                        // delta
                                        x = m_features.sp_x[i] - 8;
                                        y = m_features.sp_y[i] - 8;
                                        line = new Line2D.Double(x, y + 15, x + 8, y);
                                        g2.draw(line);
                                        line = new Line2D.Double(x + 8, y, x + 15, y + 15);
                                        g2.draw(line);
                                        line = new Line2D.Double(x + 15, y + 15, x,     y + 15);
                                        g2.draw(line);
                                        break;
                        }

                        if (Settings.useBlockedOrientation == false) return;
                        // draw block orientations
                        int bs = VeriFingerWrapper.VF_BLOCK_SIZE;
                        int hbs = bs / 2;
                        for (i = 0, yy = hbs; i < m_features.bo_height; ++i, yy += bs)
                                for (j = 0, xx = hbs; j < m_features.bo_width; ++j, xx += bs) {
                                        // convert to unsigned
                                        int dir = (int) (m_features.bo_bits[i][j] & 0xFF);
                                        if (dir != VeriFingerWrapper.VFDIR_BACKGROUND) {
                                                if (dir != VeriFingerWrapper.VFDIR_UNKNOWN) {
                                                        g2.setColor(Color.black);
                                                        g2.setStroke(new BasicStroke(1.0f));
                                                        double d = VeriFingerWrapper.VFDirToDeg(dir);
                                                        x = hbs * Math.cos(d * r);
                                                        y = hbs * Math.sin(d * r);
                                                        line = new Line2D.Double(xx, yy, xx + x, yy + y);
                                                        g2.draw(line);
                                                        line = new Line2D.Double(xx, yy, xx - x, yy - y);
                                                        g2.draw(line);
                                                } else {

                                                }
                                        }
                                }
                }
        }
}

