package neurotechnologija.utils;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Rectangle;

import javax.swing.JPanel;

/**
 * JPanel with capability to display image with simple zooming
 *
 * @see javax.swing#JPanel
 * @version 1.0 02/07/2002
 **/
public class ImagePanel extends JPanel {
        protected Image image;
        protected Dimension prefSize;
        protected double zoom;
        protected double ratio;
        protected int imageHeight, imageWidth;
        protected static final int GAP = 5; // spacing from image to panel border
        protected Color imageBackground;

        public ImagePanel() {
                this(null, 0);
        }

        public ImagePanel(Image img, double ratio) {
                super();
                imageBackground = Color.white;
                setImage(img, ratio);
        }

        /**
         * Change internal image to the specified. If <b>null</b>
         * is passed clears the internal image
         * */
        public void setImage(Image image, double ratio) {
                if (this.image != null) {
                        this.image.flush();
                        this.image = null;
                }
                // -> invariant: this.image == null
                if (image == null) return;
                this.image = image;
                this.ratio = ratio;
                zoom = 100;
                updateImageSize();
                prefSize = new Dimension(imageWidth, imageHeight);
                invalidate();
                repaint();
        }

        /**
         * @return internal image
         * */
        public Image getImage() {
                return image;
        }

        /**
         * Preferred size for this component is internal image size
         * */
        public Dimension getPreferredSize() {
                if (prefSize == null) {
                        prefSize = new Dimension(imageWidth, imageHeight);
                }
                return prefSize;
        }

        /**
         * Draw the internal image
         * */
        protected void drawImage(Graphics g) {
                Insets insets = this.getInsets();
                Rectangle bounds = this.getBounds();

                g.setColor(imageBackground);
                g.fillRect(
                        insets.left,
                        insets.top,
                        bounds.width - insets.left - insets.right,
                        bounds.height - insets.top - insets.bottom);
                if (image != null) {
                        g.setClip(
                                insets.left + GAP,
                                insets.top + GAP,
                                bounds.width - insets.left - insets.right - GAP * 2,
                                bounds.height - insets.top - insets.bottom - GAP * 2);
                        g.drawImage(
                                image,
                                insets.left + GAP,
                                insets.top + GAP,
                                imageWidth,
                                imageHeight,
                                null);
                }
        }

        /**
         * Overriden to draw image on top of component
         * */
        protected void paintComponent(Graphics g) {
                super.paintComponent(g);
                // all image drawing will be performed on copy of Graphics
                Graphics g1 = g.create();
                if (g1 != null) {
                        try {
                                drawImage(g1);
                        }
                        finally {
                                g1.dispose();
                        }
                }
        }

        private void updateImageSize() {
                imageHeight = (int) (zoom / 100.0 * image.getHeight(this) * ratio);
                imageWidth = (int) (zoom / 100.0 * image.getWidth(this) * ratio);
        }

        /**
         * Returns the imageBackground.
         * @return Color
         */
        public Color getImageBackground() {
                return imageBackground;
        }


        /**
         * Returns the zoom.
         * @return double
         */
        public double getZoom() {
                return zoom;
        }


        /**
         * Sets the imageBackground.
         * @param imageBackground The imageBackground to set
         */
        public void setImageBackground(Color imageBackground) {
                if (this.imageBackground != imageBackground) {
                        this.imageBackground = imageBackground;
                        invalidate();
                        repaint();
                }
        }


        /**
         * Sets the zoom.
         * @param zoom The zoom to set
         */
        public void setZoom(double zoom) {
                if (this.image != null && this.zoom != zoom) {
                        this.zoom = zoom;
                        updateImageSize();
                        invalidate();
                        repaint();
                }
        }


}
