unit Utils;

interface

{$WARN SYMBOL_PLATFORM OFF}

uses Windows, SysUtils, Registry;

const
  AllFilesFilter = '*.*';
  AllFilesFilterName = 'All files';

function MakeDWord(High, Low: Word): LongWord;

// Math routines
function PercentF(Count, Total: Integer): Double;
function Percent(Count, Total: Integer): Integer;

// Flag routines
function IsFlagSet(Flags, Flag: LongWord): Boolean;
procedure SetFlag(var Flags: LongWord; Flag: LongWord); overload;
procedure ClearFlag(var Flags: LongWord; Flag: LongWord);
procedure SetFlag(var Flags: LongWord; Flag: LongWord; Value: Boolean); overload;

// String routines
function IndexOf(S: string; C: Char; O: Integer = 1): Integer;

// File name routines
function MakeFullFileName(Path, Filename: string; Ext: string = ''): string;
function GetFullFileName(Filename: string): string;
function ExtractFileSubDir(Filename: string; Index: Integer): string;
function FindFile(Path: string; Filter: string = '*.*'; SearchSubdirectories: Boolean = True): string;

// File filter routines
function MakeFileFilterString(Name, Filter: string; FilterInName: Boolean = True): string;
function GetAllFilesFilterString(FilterInName: Boolean = True): string;
function FileFilterHasExtension(Filter, Ext: string): Boolean;
function ExtensionsInFileFilter(Filter: string): Integer;
function GetExtensionFromFileFilter(Filter: string; Index: Integer = 0): string;
procedure AddFileFilterString(var S: string; FilterString: string); overload;
procedure AddFileFilterString(var S: string; Name, Filter: string; FilterInName: Boolean = True); overload;
procedure AddAllFilesFilterString(var S: string; FilterInName: Boolean = True);

// Convertion to string routines
function IntToString(Value: Integer; One, Many: string; No: string = ''): string;
function BoolToString(Value: Boolean; TrueValue: string = 'True'; FalseValue: string = 'False'): string;
function TimeToString(MS: LongWord; L: Integer = 2): string; overload;
function TimeToString(Sec: Double; L: Integer = 2): string; overload;
function SizeToString(Value: LongWord; L: Integer = 2): string;

// Convertion from string routines
function StringToFloat(Value: string; E: string = ''): Double;

// Registry routines
function RegReadInteger(Reg: TRegistry; Name: string; Value: Integer): Integer;
function RegReadBool(Reg: TRegistry; Name: string; Value: Boolean): Boolean;
function RegReadFloat(Reg: TRegistry; Name: string; Value: Double): Double;
function RegReadString(Reg: TRegistry; Name: string; Value: string): string;

// High resolution timer routines
function PerformanceCounterToSeconds(TC: Int64): Double;
function GetTickCountEx: Double;

// OS version information
var
  OSMajorVersion, OSMinorVersion: LongWord;
  OSPlatformId: LongWord;
  OSBuild: LongWord;
  OSUpdateName: string;
  OSSPMajorVersion, OSSPMinorVersion: Word;
  OSSuite: Word;
  OSType: Word;
  OSName: string;

// Application version information
var
  ApplicationVersionHigh, ApplicationVersionLow: LongWord;
  ApplicationCopyright: string;
  ApplicationCompany: string;
  ApplicationRegistryKey: string;

implementation

function MakeDWord(High, Low: Word): LongWord;
begin
  Result := High shl 16 or Low;
end;

function PercentF(Count, Total: Integer): Double;
begin
  Result := 100.0 * Count / Total;
end;

function Percent(Count, Total: Integer): Integer;
begin
  Result := Round(PercentF(Count, Total));
end;

function IsFlagSet(Flags, Flag: LongWord): Boolean;
begin
  Result := (Flags and Flag) = Flag;
end;

procedure SetFlag(var Flags: LongWord; Flag: LongWord);
begin
  Flags := Flags or Flag;
end;

procedure ClearFlag(var Flags: LongWord; Flag: LongWord);
begin
  Flags := Flags and not Flag;
end;

procedure SetFlag(var Flags: LongWord; Flag: LongWord; Value: Boolean);
begin
  if Value then SetFlag(Flags, Flag) else ClearFlag(Flags, Flag);
end;

function IndexOf(S: string; C: Char; O: Integer = 1): Integer;
var
  I: Integer;
begin
  for I := O to Length(S) do
    if S[I] = C then
    begin
      Result := I;
      Exit;
    end;
  Result := -1;
end;

function MakeFullFileName(Path, Filename: string; Ext: string = ''): string;
begin
  Result := Path;
  if Path[Length(Path)] <> '\' then Result := Result + '\';
  Result := Result + Filename + Ext;
end;

function GetFullFileName(Filename: string): string;
begin
  Result := MakeFullFileName(GetCurrentDir, Filename);
end;

function ExtractFileSubDir(Filename: string; Index: Integer): string;
var
  I, J: Integer;
begin
  Result := ExtractFilePath(Filename);
  I := IndexOf(Result, ':');
  if I = -1 then I := 1 else Inc(I);
  if (I <> Length(Result) + 1) and (Result[I] = '\') then Inc(I);
  Result := Copy(Result, I, Length(Result) - I + 1);
  if Result <> EmptyStr then Result := Copy(Result, 1, Length(Result) - 1);
  if (Index = 0) or (Result = EmptyStr) then Exit;
  if Index > 0 then
  begin
    Result := Result + '\';
    J := 1;
    for I := 1 to Length(Result) do
    begin
      if Result[I] = '\' then
      begin
        Dec(Index);
        if Index = 0 then
        begin
          Result := Copy(Result, J, I - J);
          Exit;
        end
        else J := I + 1;
      end;
    end;
  end
  else begin
    Result := '\' + Result;
    J := Length(Result);
    for I := Length(Result) downto 1 do
    begin
      if Result[I] = '\' then
      begin
        Inc(Index);
        if Index = 0 then
        begin
          Result := Copy(Result, I + 1, J - I);
          Exit;
        end
        else J := I - 1;
      end;
    end;
  end;
  Result := EmptyStr;
end;

function FindFile(Path: string; Filter: string = '*.*'; SearchSubdirectories: Boolean = True): string;
var
  CurPath: string;
  FN: string;
  F: TSearchRec;
  S: string;
begin
  CurPath := Path;
  CurPath := IncludeTrailingBackslash(CurPath);
  FN := CurPath + '*';
  if FindFirst(FN, faAnyFile, F) = 0 then
  begin
    repeat begin
      if F.Attr and faDirectory = 0 then
      begin
        S := F.Name;
        if FileFilterHasExtension(Filter, ExtractFileExt(S)) then
        begin
          Result := CurPath + S;
          Break;
        end;
      end;
    end until FindNext(F) <> 0;
    FindClose(F);
  end;
  if (Result = EmptyStr) and SearchSubdirectories then
  begin
    if FindFirst(FN, faDirectory, F) = 0 then
    begin
      repeat begin
        if (F.Name <> '.') and (F.Name <> '..') then
        begin
          Result := FindFile(CurPath + F.Name, Filter, True);
          if Result <> EmptyStr then Break;
        end;
      end until FindNext(F) <> 0;
      FindClose(F);
    end;
  end;
end;

function MakeFileFilterString(Name, Filter: string; FilterInName: Boolean = True): string;
begin
  Result := Name;
  if FilterInName then Result := Result + ' (' + Filter + ')';
  Result := Result + '|' + Filter;
end;

function GetAllFilesFilterString(FilterInName: Boolean = True): string;
begin
  Result := MakeFileFilterString(AllFilesFilterName, AllFilesFilter, FilterInName);
end;

function FileFilterHasExtension(Filter, Ext: string): Boolean;
var
  I: Integer;
  E: string;
  L, S: Integer;
begin
  for I := Length(Ext) downto 1 do
    if Ext[I] = '.' then Break
    else E := Ext[I] + E;
  L := Length(Filter);
  S := 0;
  for I := 1 to L + 1 do
    if (S <> 0) and ((I = L + 1) or (Filter[I] = ';')) then
    begin
      if Copy(Filter, S, I - S) = E then
      begin
        Result := True;
        Exit;
      end;
      S := 0;
    end
    else if (I <> L + 1) and (Filter[I] = '.') then S := I + 1;
  Result := False;
end;

function ExtensionsInFileFilter(Filter: string): Integer;
var
  I: Integer;
  S: Boolean;
  L: Integer;
begin
  L := Length(Filter);
  S := False;
  Result := 0;
  for I := 1 to L + 1 do
    if S and ((I = L + 1) or (Filter[I] = ';')) then
    begin
      Inc(Result);
      S := False;
    end
    else if (I <> L + 1) and (Filter[I] = '.') then S := True;
end;

function GetExtensionFromFileFilter(Filter: string; Index: Integer = 0): string;
var
  I: Integer;
  S, L: Integer;
begin
  L := Length(Filter);
  S := 0;
  for I := 1 to L + 1 do
    if (S <> 0) and ((I = L + 1) or (Filter[I] = ';')) then
    begin
      if Index = 0 then
      begin
        Result := Copy(Filter, S, I - S);
        Exit;
      end;
      Dec(Index);
      S := 0;
    end
    else if (I <> L + 1) and (Filter[I] = '.') then S := I + 1;
  Result := EmptyStr;
end;

procedure AddFileFilterString(var S: string; FilterString: string);
begin
  if S <> EmptyStr then S := S + '|';
  S := S + FilterString;
end;

procedure AddFileFilterString(var S: string; Name, Filter: string; FilterInName: Boolean = True);
begin
  AddFileFilterString(S, MakeFileFilterString(Name, Filter, FilterInName));
end;

procedure AddAllFilesFilterString(var S: string; FilterInName: Boolean = True);
begin
  AddFileFilterString(S, GetAllFilesFilterString(FilterInName));
end;

function IntToString(Value: Integer; One, Many: string; No: string = ''): string;
begin
  if (Value = 0) and (No <> EmptyStr) then Result := No else Result := IntToStr(Value);
  Result := Result + ' ';
  if (Value mod 10 = 1) and (Value mod 100 <> 11) then Result := Result + One else Result := Result + Many;
end;

function BoolToString(Value: Boolean; TrueValue: string = 'True'; FalseValue: string = 'False'): string;
begin
  if Value then Result := TrueValue
  else Result := FalseValue;
end;

function TimeToString(MS: LongWord; L: Integer = 2): string;
var
  I: Integer;
begin
  Result := IntToStr(MS mod 1000) + ' ms';
  MS := MS div 1000;
  if MS <> 0 then
  begin
    Result := IntToStr(MS mod 60) + ' s ' + Result;
    MS := MS div 60;
    if MS <> 0 then
    begin
      Result := IntToStr(MS mod 60) + ' m ' + Result;
      MS := MS div 60;
      if MS <> 0 then
      begin
        Result := IntToStr(MS mod 60) + ' h ' + Result;
        MS := MS div 24;
        if MS <> 0 then
        begin
          Result := IntToStr(MS) + ' d ' + Result;
        end;
      end;
    end;
  end;
  if L > 0 then
  begin
    I := 1;
    while (L <> 0) and (I <> -1) do
    begin
      I := IndexOf(Result, ' ', I + 1);
      if I <> -1 then I := IndexOf(Result, ' ', I + 1)
      else Break;
      Dec(L);
    end;
    if I = -1 then I := Length(Result) + 1;
    Result := Copy(Result, 1, I - 1);
  end;
end;

function TimeToString(Sec: Double; L: Integer = 2): string;
var
  Add: Boolean;
  I: Integer;
begin
  Sec := Sec * 1000;
  Add := True;
  if Sec >= 1 then
  begin
    Result := TimeToString(LongWord(Round(Sec)), L);
    if L > 0 then
    begin
      I := 0;
      while (L <> 0) and (I <> -1) do
      begin
        I := IndexOf(Result, ' ', I + 1);
        if I <> -1 then I := IndexOf(Result, ' ', I + 1)
        else Break;
        Dec(L);
      end;
    end;
    Add := L <> 0;
  end;
  if Add then
  begin
    if Result <> EmptyStr then Result := Result + ' ';
    Sec := Sec - Trunc(sec);
    Sec := Sec * 1000;
    Result := Result + IntToStr(Round(Sec)) + ' mks';
  end;
end;

function SizeToString(Value: LongWord; L: Integer = 2): string;
var
  I: Integer;
begin
  Result := IntToString(Value mod 1024, 'byte', 'bytes');
  Value := Value div 1024;
  if Value <> 0 then
  begin
    Result := IntToStr(Value mod 1024) + ' KB ' + Result;
    Value := Value div 1024;
    if Value <> 0 then
    begin
      Result := IntToStr(Value mod 1024) + ' MB ' + Result;
      Value := Value div 1024;
      if Value <> 0 then
      begin
        Result := IntToStr(Value) + ' GB ' + Result;
      end;
    end;
  end;
  if L > 0 then
  begin
    I := 1;
    while (L <> 0) and (I <> -1) do
    begin
      I := IndexOf(Result, ' ', I + 1);
      if I <> -1 then I := IndexOf(Result, ' ', I + 1)
      else Break;
      Dec(L);
    end;
    if I = -1 then I := Length(Result) + 1;
    Result := Copy(Result, 1, I - 1);
  end;
end;

function StringToFloat(Value: string; E: string = ''): Double;
var
  I: Integer;
  L: Integer;
  Dot: Boolean;
  Exp: Boolean;
  EE: string;
begin
  Value := Trim(Value);
  L := Length(Value);
  if (L <> 0) and (Value[1] = '+') or (Value[1] = '-') then I := 2 else I := 1;
  Dot := False;
  Exp := False;
  while I <= L do
  begin
    if (Value[I] = DecimalSeparator) then
      if Dot then Break
      else Dot := True
    else if (Value[I] = 'E') or (Value[I] = 'e') then
      if Exp then Break
      else begin
        Exp := True;
        Dot := True;
        if I < L then if (Value[I + 1] = '+') or (Value[I + 1] = '-') then Inc(I);
      end
    else if (Value[I] < '0') or (Value[I] > '9') then Break;
    Inc(I);
  end;
  if I <= L then
  begin
    EE := Copy(Value, I, L - I + 1);
    EE := Trim(EE);
  end
  else EE := EmptyStr;
  E := Trim(E);
  if (Value <> EmptyStr) and ((EE = EmptyStr) or (EE = E)) then
  begin
    Result := StrToFloat(Copy(Value, 1, I - 1));
    Exit;
  end;
  raise Exception.Create('Can not convert string to float');
end;

function RegReadInteger(Reg: TRegistry; Name: string; Value: Integer): Integer;
begin
  try
    Result := Reg.ReadInteger(Name);
  except
    Result := Value;
  end;
end;

function RegReadBool(Reg: TRegistry; Name: string; Value: Boolean): Boolean;
begin
  try
    Result := Reg.ReadBool(Name);
  except
    Result := Value;
  end;
end;

function RegReadFloat(Reg: TRegistry; Name: string; Value: Double): Double;
var
  T: LongWord;
  S: LongWord;
begin
  try
    Result := Reg.ReadFloat(Name);
  except
    // Try QWORD read
    try
      S := SizeOf(Double);
      RegQueryValueEx(Reg.CurrentKey, PChar(Name), nil, @T, PByte(@Result), @S);
      if (T <> 11 {REG_QWORD}) or (S <> SizeOf(Double)) then raise Exception.Create(''); 
    except
      Result := Value;
    end;
  end;
end;

function RegReadString(Reg: TRegistry; Name: string; Value: string): string;
begin
  try
    Result := Reg.ReadString(Name);
  except
    Result := Value;
  end;
end;

var
  PCF: Int64;

function PerformanceCounterToSeconds(TC: Int64): Double;
begin
  Result := TC / PCF;
end;

function GetTickCountEx: Double;
var
  TC: Int64;
begin
  if not QueryPerformanceCounter(TC) then raise Exception.Create('Can not query performance counter');
  Result := PerformanceCounterToSeconds(TC);
end;

type
  POSVersionInfoExA = ^TOSVersionInfoExA;
  POSVersionInfoExW = ^TOSVersionInfoExW;
  POSVersionInfoEx = POSVersionInfoExA;
  _OSVERSIONINFOEXA = record
    dwOSVersionInfoSize: DWORD;
    dwMajorVersion: DWORD;
    dwMinorVersion: DWORD;
    dwBuildNumber: DWORD;
    dwPlatformId: DWORD;
    szCSDVersion: array[0..127] of AnsiChar; { Maintenance string for PSS usage }
    wServicePackMajor: Word;
    wServicePackMinor: Word;
    wSuiteMask: Word;
    wProductType: Byte;
    wReserved: Byte;
  end;
  _OSVERSIONINFOEXW = record
    dwOSVersionInfoSize: DWORD;
    dwMajorVersion: DWORD;
    dwMinorVersion: DWORD;
    dwBuildNumber: DWORD;
    dwPlatformId: DWORD;
    szCSDVersion: array[0..127] of WideChar; { Maintenance string for PSS usage }
    wServicePackMajor: Word;
    wServicePackMinor: Word;
    wSuiteMask: Word;
    wProductType: Byte;
    wReserved: Byte;
  end;
  _OSVERSIONINFOEX = _OSVERSIONINFOEXA;
  TOSVersionInfoExA = _OSVERSIONINFOEXA;
  TOSVersionInfoExW = _OSVERSIONINFOEXW;
  TOSVersionInfoEx = TOSVersionInfoExA;
  OSVERSIONINFOEXA = _OSVERSIONINFOEXA;
  OSVERSIONINFOEXW = _OSVERSIONINFOEXW;
  OSVERSIONINFOEX = OSVERSIONINFOEXA;

function GetVersionEx(lpVersionInformation: POSVersionInfo): BOOL; stdcall; external kernel32 name 'GetVersionExA';
function GetVersionExA(lpVersionInformation: POSVersionInfo): BOOL; stdcall; external kernel32 name 'GetVersionExA';
function GetVersionExW(lpVersionInformation: POSVersionInfo): BOOL; stdcall; external kernel32 name 'GetVersionExW';

const
  VER_NT_WORKSTATION       = $0000001;
  VER_NT_DOMAIN_CONTROLLER = $0000002;
  VER_NT_SERVER            = $0000003;

  VER_SERVER_NT                      = $80000000;
  VER_WORKSTATION_NT                 = $40000000;
  VER_SUITE_SMALLBUSINESS            = $00000001;
  VER_SUITE_ENTERPRISE               = $00000002;
  VER_SUITE_BACKOFFICE               = $00000004;
  VER_SUITE_COMMUNICATIONS           = $00000008;
  VER_SUITE_TERMINAL                 = $00000010;
  VER_SUITE_SMALLBUSINESS_RESTRICTED = $00000020;
  VER_SUITE_EMBEDDEDNT               = $00000040;
  VER_SUITE_DATACENTER               = $00000080;
  VER_SUITE_SINGLEUSERTS             = $00000100;
  VER_SUITE_PERSONAL                 = $00000200;
  VER_SUITE_BLADE                    = $00000400;

procedure RetrieveOSInfo;
var
  VIE: TOSVersionInfoEx;
  VI: TOSVersionInfo;
begin
  VIE.dwOSVersionInfoSize := SizeOf(VIE);
  if GetVersionEx(POSVersionInfo(@VIE)) then
  begin
    OSMajorVersion := VIE.dwMajorVersion;
    OSMinorVersion := VIE.dwMinorVersion;
    OSPlatformId := VIE.dwPlatformId;
    OSBuild := VIE.dwBuildNumber;
    OSUpdateName := VIE.szCSDVersion;
    OSSPMajorVersion := VIE.wServicePackMajor;
    OSSPMinorVersion := VIE.wServicePackMinor;
    OSSuite := VIE.wSuiteMask;
    OSType := VIE.wProductType;
  end
  else begin
    OSSPMajorVersion := 0;
    OSSPMinorVersion := 0;
    OSSuite := 0;
    OSType := 0;
    VI.dwOSVersionInfoSize := SizeOf(VI);
    if GetVersionEx(@VI) then
    begin
      OSMajorVersion := VI.dwMajorVersion;
      OSMinorVersion := VI.dwMinorVersion;
      OSPlatformId := VI.dwPlatformId;
      OSBuild := VI.dwBuildNumber;
      OSUpdateName := VI.szCSDVersion;
    end
    else begin
      OSMajorVersion := 0;
      OSMinorVersion := 0;
      OSPlatformId := 0;
      OSBuild := 0;
    end;
  end;
  if OSPlatformId <> VER_PLATFORM_WIN32_NT then OSBuild := LoWord(OSBuild);

  if OSPlatformId = VER_PLATFORM_WIN32s then OSName := 'Win32s'
  else if OSPlatformId = VER_PLATFORM_WIN32_WINDOWS then
  begin
    if OSMajorVersion > 4 then OSName := 'Windows'
    else if OSMajorVersion = 4 then
    begin
      if OSMinorVersion >= 90 then OSName := 'Windows Me'
      else if OSMinorVersion >= 10 then OSName := 'Windows 98'
      else OSName := 'Windows 95';
    end
    else OSName := Format('Windows %ul.%ul', [OSMajorVersion, OSMinorVersion]);
  end
  else if OSPlatformId = VER_PLATFORM_WIN32_NT then
  begin
    if OSMajorVersion > 5 then OSName := 'Windows NT'
    else if OSMajorVersion = 5 then
    begin
      if OSMinorVersion >= 1 then
      begin
        if OSType = VER_NT_SERVER then
        begin
          if OSSuite and VER_SUITE_ENTERPRISE <> 0 then OSName := 'Windows .NET Enterprise Server'
          else OSName := 'Windows .NET Server';
        end
        else if OSType = VER_NT_WORKSTATION then
        begin
          if OSSuite and VER_SUITE_PERSONAL <> 0 then OSName := 'Windows XP Home Edition'
          else OSName := 'Windows XP Professional';
        end
        else begin
          OSName := 'Windows XP/.NET';
        end;
      end
      else begin
        if OSType = VER_NT_SERVER then
        begin
          if OSSuite and VER_SUITE_DATACENTER <> 0 then OSName := 'Windows 2000 DataCenter Server'
          else if OSSuite and VER_SUITE_ENTERPRISE <> 0 then OSName := 'Windows 2000 Advanced Server'
          else OSName := 'Windows 2000 Server';
        end
        else if OSType = VER_NT_WORKSTATION then OSName := 'Windows 2000 Professional'
          else OSName := 'Windows 2000';
      end;
    end
    else if OSMajorVersion = 4 then
    begin
      if OSType = VER_NT_SERVER then OSName := 'Windows NT 4.0 Server'
      else if OSType = VER_NT_WORKSTATION then OSName := 'Windows NT 4.0 Workstation'
      else OSName := 'Windows NT 4.0';
    end
    else OSName := Format('Windows NT %ul.%ul', [OSMajorVersion, OSMinorVersion]);
  end
  else OSName := '[Unknown]';
  if OSUpdateName <> EmptyStr then
    if OSPlatformId = VER_PLATFORM_WIN32_WINDOWS then OSName := OSName + OSUpdateName
    else OSName := OSName + ', ' + OSUpdateName;
end;

initialization
  if not QueryPerformanceFrequency(PCF) then raise Exception.Create('Can not query performance frequency');
  RetrieveOSInfo;
  ApplicationVersionHigh := MakeLong(0, 1);
  ApplicationVersionLow := MakeLong(0, 0);
  ApplicationCopyright := EmptyStr;
  ApplicationCompany := EmptyStr;
  ApplicationRegistryKey := EmptyStr;

end.

