unit VFImage;

interface

uses Classes, SysUtils, Windows, Graphics, Utils;

type
  TVFImage = class
  private
    FWidth: Integer;
    FHeight: Integer;
    FSize: LongWord;
    FBits: PByte;

    procedure CheckPixelIndex(X, Y: Integer);

    function GetIsEmpty: Boolean;
    function GetPixel(X, Y: Integer): Byte;
    procedure SetHeight(const Value: Integer);
    procedure SetPixel(X, Y: Integer; const Value: Byte);
    procedure SetWidth(const Value: Integer);

  public
    constructor Create;
    constructor CreateSized(Width, Height: Integer);
    constructor CreateFromFile(FileName: string);
    destructor Destroy; override;

    procedure Assign(Image: TVFImage); overload;
    procedure Assign(Width, Height: Integer; Bits: PByte); overload;
    procedure Clear;
    procedure SetSize(NewWidth, NewHeight: Integer);

    function LoadFromFile(Filename: string): Boolean;
    function SaveToFile(Filename: string; FormatIndex: Integer = -1): Boolean;

    function GetBitmap(ColorMin: TColor = clBlack; ColorMax: TColor = clWhite): HBITMAP;

    property Width: Integer read FWidth write SetWidth;
    property Height: Integer read FHeight write SetHeight;
    property Bits: PByte read FBits;
    property IsEmpty: Boolean read GetIsEmpty;

    property Pixels[X, Y: Integer]: Byte read GetPixel write SetPixel;

  end;

  TVFImageFormat = class
  protected
    function GetCanLoad: Boolean; virtual; abstract;
    function GetCanSave: Boolean; virtual; abstract;
    function GetFileFilter: string; virtual; abstract;
    function GetFileFilterName: string; virtual; abstract;

    procedure Load(Image: TVFImage; Stream: TStream); overload; virtual;
    procedure Save(Image: TVFImage; Stream: TStream); overload; virtual;

  public
    constructor Create;
    destructor Destroy; override;

    function GetFileFilterString(FilterInName: Boolean = True): string;

    function Load(Image: TVFImage; Filename: string): Boolean; overload; virtual;
    function Save(Image: TVFImage; Filename: string): Boolean; overload; virtual;

    property CanLoad: Boolean read GetCanLoad;
    property CanSave: Boolean read GetCanSave;
    property FileFilter: string read GetFileFilter;
    property FileFilterName: string read GetFileFilterName;

  end;

  TVFImageIO = class
  private
    FFormats: TList;

    procedure AddFormat(Format: TVFImageFormat);
    procedure RemoveFormat(Format: TVFImageFormat);

    function GetCanLoad: Boolean;
    function GetCanSave: Boolean;
    function GetFormat(Index: Integer): TVFImageFormat;
    function GetFormatCount: Integer;
    function GetLoadFileFilter: string;
    function GetSaveFileFilter: string;

  public
    constructor Create;
    destructor Destroy; override;

    function GetLoadFileFilterString(FilterInName: Boolean = True; AllFilters: Boolean = True; AllFiles: Boolean = true): string;
    function GetSaveFileFilterString(FilterInName: Boolean = True): string;

    function Load(Image: TVFImage; Filename: string): Boolean;
    function Save(Image: TVFImage; Filename: string; FormatIndex: Integer = -1): Boolean;

    property FormatCount: Integer read GetFormatCount;
    property Formats[Index: Integer]: TVFImageFormat read GetFormat;

    property CanLoad: Boolean read GetCanLoad;
    property CanSave: Boolean read GetCanSave;

    property LoadFileFilter: string read GetLoadFileFilter;
    property SaveFileFilter: string read GetSaveFileFilter;

  end;

var
  ImageIO: TVFImageIO;

implementation

{ TVFImage }

procedure TVFImage.Assign(Image: TVFImage);
begin
  Assign(Image.FWidth, Image.FHeight, Image.FBits);
end;

procedure TVFImage.Assign(Width, Height: Integer; Bits: PByte);
begin
  SetSize(Width, Height);
  if FSize <> 0 then Move(Bits^, FBits^, FSize);
end;

procedure TVFImage.CheckPixelIndex(X, Y: Integer);
begin
  if (X < 0) or (X >= FWidth) or (Y < 0) or (Y >= FHeight) then
    raise Exception.Create('Image pixel index out of bounds');
end;

procedure TVFImage.Clear;
begin
  if not IsEmpty then
  begin
    FWidth := 0;
    FHeight := 0;
    FSize := 0;
    FreeMem(FBits);
    FBits := nil;
  end;
end;

constructor TVFImage.Create;
begin
  FWidth := 0;
  FHeight := 0;
  FBits := nil;
  FSize := 0;
end;

constructor TVFImage.CreateSized(Width, Height: Integer);
begin
  FWidth := 0;
  FHeight := 0;
  FBits := nil;
  FSize := 0;
  SetSize(Width, Height);
end;

constructor TVFImage.CreateFromFile(FileName: string);
begin
  FWidth := 0;
  FHeight := 0;
  FBits := nil;
  FSize := 0;
  LoadFromFile(FileName);
end;

destructor TVFImage.Destroy;
begin
  Clear;
  inherited;
end;

function TVFImage.GetBitmap(ColorMin: TColor; ColorMax: TColor): HBITMAP;
var
  BI: PBitmapInfo;
  RGBMin, RGBMax: COLORREF;
  BlueMin, GreenMin, RedMin, BlueMax, GreenMax, RedMax, BlueDelta, GreenDelta, RedDelta: Integer;
  I: Integer;
  Bits: Pointer;
  WidthBytes: LongWord;
  DP, SP: LongWord;
begin
  if IsEmpty then
  begin
    Result := 0;
    Exit;
  end;
  BI := PBitmapInfo(AllocMem(SizeOf(TBitmapInfoHeader) + SizeOf(TRGBQuad) * 256));
  BI.bmiHeader.biBitCount := 8;
  BI.bmiHeader.biClrImportant := 0;
  BI.bmiHeader.biClrUsed := 0;
  BI.bmiHeader.biCompression := BI_RGB;
  BI.bmiHeader.biPlanes := 1;
  BI.bmiHeader.biSize := SizeOf(TBitmapInfoHeader);
  BI.bmiHeader.biXPelsPerMeter := 0;
  BI.bmiHeader.biYPelsPerMeter := 0;
  BI.bmiHeader.biHeight := FHeight;
  BI.bmiHeader.biWidth := FWidth;
  RGBMin := ColorToRGB(ColorMin);
  RGBMax := ColorToRGB(ColorMax);
  BlueMin := GetBValue(RGBMin);
  GreenMin := GetGValue(RGBMin);
  RedMin := GetRValue(RGBMin);
  BlueMax := GetBValue(RGBMax);
  GreenMax := GetGValue(RGBMax);
  RedMax := GetRValue(RGBMax);
  BlueDelta := BlueMax - BlueMin;
  GreenDelta := GreenMax - GreenMin;
  RedDelta := RedMax - RedMin;
{$R-}
  for I := 0 to 255 do
  begin
    BI.bmiColors[i].rgbBlue := Byte(I * BlueDelta div 255 + BlueMin);
    BI.bmiColors[i].rgbGreen := Byte(I * GreenDelta div 255 + GreenMin);
    BI.bmiColors[i].rgbRed := Byte(I * RedDelta div 255 + RedMin);
    BI.bmiColors[i].rgbReserved := 0;
  end;
  Result := CreateDIBSection(0, BI^, DIB_RGB_COLORS, Bits, 0, 0);
  if Result <> 0 then
  begin
    WidthBytes := Width div 4;
    if Width mod 4 <> 0 then Inc(WidthBytes);
    WidthBytes := WidthBytes * 4;
    DP := LongWord(Bits);
    SP := LongWord(FBits) + LongWord((FHeight - 1) * FWidth);
    for I := 0 to FHeight - 1 do
    begin
      Move(PByte(SP)^, PByte(DP)^, FWidth);
      Inc(DP, WidthBytes);
      Dec(SP, FWidth);
    end;
  end;
  FreeMem(BI);
end;

function TVFImage.GetIsEmpty: Boolean;
begin
  Result := FBits = nil;
end;

function TVFImage.GetPixel(X, Y: Integer): Byte;
begin
  CheckPixelIndex(X, Y);
  Result := PByte(LongWord(FBits) + LongWord(Y) * LongWord(FWidth) + LongWord(X))^;
end;

function TVFImage.LoadFromFile(Filename: string): Boolean;
begin
  Result := ImageIO.Load(Self, Filename);
end;

function TVFImage.SaveToFile(Filename: string; FormatIndex: Integer): Boolean;
begin
  Result := ImageIO.Save(Self, Filename, FormatIndex);
end;

procedure TVFImage.SetHeight(const Value: Integer);
begin
  SetSize(FWidth, Value);
end;

procedure TVFImage.SetPixel(X, Y: Integer; const Value: Byte);
begin
  CheckPixelIndex(X, Y);
  PByte(LongWord(FBits) + LongWord(Y) * LongWord(FWidth) + LongWord(X))^ := Value;
end;

procedure TVFImage.SetSize(NewWidth, NewHeight: Integer);
var
  NewSize: LongWord;
begin
  if (NewWidth <= 0) or (NewHeight <= 0) then
  begin
    NewWidth := 0;
    NewHeight := 0;
  end;
  NewSize := NewWidth * NewHeight;
  if NewSize <> FSize then
  begin
    ReAllocMem(FBits, NewSize);
    FSize := NewSize;
  end;
  FWidth := NewWidth;
  FHeight := NewHeight;
end;

procedure TVFImage.SetWidth(const Value: Integer);
begin
  SetSize(Value, FHeight);
end;

{ TVFImageFormat }

constructor TVFImageFormat.Create;
begin
  ImageIO.AddFormat(Self);
end;

destructor TVFImageFormat.Destroy;
begin
  ImageIO.RemoveFormat(Self);
  inherited;
end;

function TVFImageFormat.GetFileFilterString(FilterInName: Boolean): string;
begin
  Result := MakeFileFilterString(FileFilterName, FileFilter, FilterInName);
end;

function TVFImageFormat.Load(Image: TVFImage; Filename: string): Boolean;
var
  Stream: TStream;
begin
  Stream := nil;
  try
    Stream := TFileStream.Create(Filename, fmOpenRead or fmShareDenyWrite);
    Load(Image, Stream);
    Result := True;
  except
    on E: Exception do begin
      Image.Clear;
      Result := False;
    end;
  end;
  Stream.Free;
end;

procedure TVFImageFormat.Load(Image: TVFImage; Stream: TStream);
begin
  raise Exception.Create('TVFImageFormat.Load not implemented');
end;

function TVFImageFormat.Save(Image: TVFImage; Filename: string): Boolean;
var
  Stream: TStream;
begin
  Stream := nil;
  try
    Stream := TFileStream.Create(Filename, fmCreate or fmShareDenyWrite);
    Save(Image, Stream);
    Result := True;
  except
    on E: Exception do Result := False;
  end;
  Stream.Free;
end;

procedure TVFImageFormat.Save(Image: TVFImage; Stream: TStream);
begin
  raise Exception.Create('TVFImageFormat.Save not implemented');
end;

{ TVFImageIO }

procedure TVFImageIO.AddFormat(Format: TVFImageFormat);
begin
  FFormats.Add(Format)
end;

constructor TVFImageIO.Create;
begin
  FFormats := TList.Create;
end;

destructor TVFImageIO.Destroy;
begin
  FFormats.Free;
  inherited;
end;

function TVFImageIO.GetCanLoad: Boolean;
var
  I: Integer;
begin
  for I := 0 to FormatCount - 1 do
    if Formats[I].CanLoad then
    begin
      Result := True;
      Exit;
    end;
  Result := False;
end;

function TVFImageIO.GetCanSave: Boolean;
var
  I: Integer;
begin
  for I := 0 to FormatCount - 1 do
    if Formats[I].CanSave then
    begin
      Result := True;
      Exit;
    end;
  Result := False;
end;

function TVFImageIO.GetFormat(Index: Integer): TVFImageFormat;
begin
  if (Index < 0) or (Index >= FFormats.Count) then
    raise Exception.Create('ImageIO format index out of bounds');
  Result := TVFImageFormat(FFormats[Index]);
end;

function TVFImageIO.GetFormatCount: Integer;
begin
  Result := FFormats.Count;
end;

function TVFImageIO.GetLoadFileFilter: string;
var
  I: Integer;
begin
  for I := 0 to FormatCount - 1 do
    if Formats[I].CanLoad then
    begin
      if Result <> EmptyStr then Result := Result + ';';
      Result := Result + Formats[I].FileFilter;
    end;
end;

function TVFImageIO.GetLoadFileFilterString(FilterInName, AllFilters,
  AllFiles: Boolean): string;
var
  AllFilter: string;
  I: Integer;
begin
  if AllFilters then
  begin
    AllFilter := LoadFileFilter;
    if AllFilter <> EmptyStr then AddFileFilterString(Result, 'All supported', AllFilter, FilterInName);
  end;
  for I := 0 to FormatCount - 1 do
    if Formats[I].CanLoad then
      AddFileFilterString(Result, Formats[I].GetFileFilterString(FilterInName));
  if AllFiles then
    AddAllFilesFilterString(Result, FilterInName);
end;

function TVFImageIO.GetSaveFileFilter: string;
var
  I: Integer;
begin
  for I := 0 to FormatCount - 1 do
    if Formats[I].CanSave then
    begin
      if Result <> EmptyStr then Result := Result + ';';
      Result := Result + Formats[I].FileFilter;
    end;
end;

function TVFImageIO.GetSaveFileFilterString(FilterInName: Boolean): string;
var
  I: Integer;
begin
  for I := 0 to FormatCount - 1 do
    if Formats[I].CanSave then
      AddFileFilterString(Result, Formats[I].GetFileFilterString(FilterInName));
end;

function TVFImageIO.Load(Image: TVFImage; Filename: string): Boolean;
var
  I: Integer;
begin
  for I := 0 to FormatCount - 1 do
    if Formats[I].CanLoad and Formats[I].Load(Image, Filename) then
    begin
      Result := True;
      Exit;
    end;
  Result := False;
end;

procedure TVFImageIO.RemoveFormat(Format: TVFImageFormat);
begin
  FFormats.Remove(Format);
end;

function TVFImageIO.Save(Image: TVFImage; Filename: string; FormatIndex: Integer): Boolean;
var
  Ext: string;
  I: Integer;
begin
  if Image.IsEmpty then
  begin
    Result := False;
    Exit;
  end;
  Ext := ExtractFileExt(Filename);
  if FormatIndex = -1 then
  begin
    if Ext <> EmptyStr then
    begin
      for I := 0 to FormatCount - 1 do
        if Formats[I].CanSave and FileFilterHasExtension(Formats[I].FileFilter, Ext) then
        begin
          FormatIndex := I;
          Break;
        end;
    end;
  end;
  if FormatIndex = -1 then
  begin
    Result := False;
    Exit;
  end;
  if Ext = EmptyStr then
    Filename := ChangeFileExt(Filename, '.' + GetExtensionFromFileFilter(Formats[FormatIndex].FileFilter));
  Result := Formats[FormatIndex].CanSave and Formats[FormatIndex].Save(Image, Filename);
end;

initialization
  ImageIO := TVFImageIO.Create;

finalization
  ImageIO.Free;

end.

