unit VFOptions;

interface

uses Windows, Registry, Utils, VFinger, VFFileList;

type
  TVFDemoMode = (vfdmNone = -1, vfdmEnrollment, vfdmEnrollmentWithGen, vfdmVerification, vfdmIdentification);

  TVFOptions = record
    StartupMode: TVFDemoMode;

    VeriFingerMode: Integer;

    UseLowQualityImages: Boolean;

    FileResolution: Integer;
    FileFingerID: string;
    FilePrompt: Boolean;

    ScannerFingerID: string;
    ScannerPrompt: Boolean;

    CheckMinMinutiaCount: Boolean;
    MinMinutiaCount: Integer;
    SearchForDuplicates: Boolean;
    WarnDifferentFingerID: Boolean;

    UseG: Boolean;
    DatabasePercent: Integer;
    MatchAll: Boolean;
    ShowAllResults: Boolean;

    ReturnedImage: Integer;

    FARate: Double;
    MatchingSpeed: Integer;
    MaximalRotation: Integer;

    GenFARate: Double;
    GenMaximalRotation: Integer;
  end;

function VFDemoModeToString(Value: TVFDemoMode): string;

procedure InitDefaultOptions;
procedure SynchronizeOptions(var Options: TVFOptions);
procedure UpdateOptions(var Options: TVFOptions);

procedure SaveOptions(var Options: TVFOptions; RegistryKey: string; RootKey: HKEY = HKEY_CURRENT_USER);
procedure LoadOptions(var Options: TVFOptions; RegistryKey: string; RootKey: HKEY = HKEY_CURRENT_USER);

var
  DefaultOptions: TVFOptions;

implementation

function VFDemoModeToString(Value: TVFDemoMode): string;
begin
  case Value of
    vfdmNone: Result := 'None';
    vfdmEnrollment: Result := 'Enrollment';
    vfdmEnrollmentWithGen: Result := 'Enrollment with generalization';
    vfdmIdentification: Result := 'Identification';
    vfdmVerification: Result := 'Verification';
  else
    Result := '[unknown mode]';
  end;
end;

procedure InitDefaultOptions;
begin
  with DefaultOptions do
  begin
    StartupMode := vfdmEnrollment;

    UseLowQualityImages := True;

    FileResolution := VF_IMAGE_RESOLUTION;
    FileFingerID := DefaultFileFingerID;
    FilePrompt := True;

    ScannerFingerID := DefaultScannerFingerID;
    ScannerPrompt := False;

    CheckMinMinutiaCount := True;
    MinMinutiaCount := 10;
    SearchForDuplicates := True;
    WarnDifferentFingerID := True;

    UseG := True;
    DatabasePercent := 100;
    MatchAll := False;
    ShowAllResults := False;
  end;
  SynchronizeOptions(DefaultOptions);
end;

procedure SynchronizeOptions(var Options: TVFOptions);
begin
  with Options do
  begin
    VeriFingerMode := VFGetIntParameter(VFP_MODE);

    ReturnedImage := VFGetIntParameter(VFP_RETURNED_IMAGE);

    FARate := VFMatchingThresholdToFAR(VFGetIntParameter(VFP_MATCHING_THRESHOLD));
    MatchingSpeed := VFGetIntParameter(VFP_MATCHING_SPEED);
    MaximalRotation := VFGetIntParameter(VFP_MAXIMAL_ROTATION);

    GenFARate := VFMatchingThresholdToFAR(VFGetIntParameter(VFP_GENERALIZATION_THRESHOLD));
    GenMaximalRotation := VFGetIntParameter(VFP_GEN_MAXIMAL_ROTATION);
  end;
end;

procedure UpdateOptions(var Options: TVFOptions);
begin
  with Options do
  begin
    try
      VFSetIntParameter(VFP_MODE, VeriFingerMode);
    except
    end;

    try
      VFSetIntParameter(VFP_RETURNED_IMAGE, ReturnedImage);
    except
    end;

    try
      VFSetIntParameter(VFP_MATCHING_THRESHOLD, VFFARToMatchingThreshold(FARate));
    except
    end;
    try
      VFSetIntParameter(VFP_MATCHING_SPEED, MatchingSpeed);
    except
    end;
    try
      VFSetIntParameter(VFP_MAXIMAL_ROTATION, MaximalRotation);
    except
    end;

    try
      VFSetIntParameter(VFP_GENERALIZATION_THRESHOLD, VFFARToMatchingThreshold(GenFARate));
    except
    end;
    try
      VFSetIntParameter(VFP_GEN_MAXIMAL_ROTATION, GenMaximalRotation);
    except
    end;
  end;
  SynchronizeOptions(Options);
end;

procedure SaveOptions(var Options: TVFOptions; RegistryKey: string; RootKey: HKEY = HKEY_CURRENT_USER);
var
  Reg: TRegistry;
begin
  Reg := TRegistry.Create;
  Reg.RootKey := RootKey;
  Reg.OpenKey(RegistryKey, True);
  try
    with Options do
    begin
      Reg.WriteInteger('Startup mode', Integer(StartupMode));

      Reg.WriteInteger('VeriFinger mode', VeriFingerMode);

      Reg.WriteBool('Use low quality images', UseLowQualityImages);

      Reg.WriteInteger('File resolution', FileResolution);
      Reg.WriteString('File finger ID', FileFingerID);
      Reg.WriteBool('File info prompt', FilePrompt);

      Reg.WriteString('Scanner finger ID', ScannerFingerID);
      Reg.WriteBool('Scanner info prompt', ScannerPrompt);

      Reg.WriteBool('Check min minutia count', CheckMinMinutiaCount);
      Reg.WriteInteger('Min minutia count', MinMinutiaCount);
      Reg.WriteBool('Search for duplicates', SearchForDuplicates);
      Reg.WriteBool('Warn different finger id', WarnDifferentFingerID);

      Reg.WriteBool('Use G', Options.UseG);
      Reg.WriteInteger('Database percent', DatabasePercent);
      Reg.WriteBool('Match all database', MatchAll);
      Reg.WriteBool('Show all results', ShowAllResults);

      Reg.WriteInteger('Returned image', ReturnedImage);

      Reg.WriteFloat('FA rate', FARate);
      Reg.WriteInteger('Matching speed', MatchingSpeed);
      Reg.WriteInteger('Maximal rotation', MaximalRotation);

      Reg.WriteFloat('Gen FA rate', GenFARate);
      Reg.WriteInteger('Gen maximal rotation', GenMaximalRotation);
    end;
  finally
    Reg.CloseKey;
    Reg.Free;
  end;
end;

procedure LoadOptions(var Options: TVFOptions; RegistryKey: string; RootKey: HKEY = HKEY_CURRENT_USER);
var
  Reg: TRegistry;
begin
  Reg := TRegistry.Create;
  Reg.RootKey := RootKey;
  try
    Reg.OpenKeyReadOnly(RegistryKey);
  except
    Reg.Free;
    Options := DefaultOptions;
    Exit;
  end;
  try
    with Options do
    begin
      StartupMode := TVFDemoMode(RegReadInteger(Reg, 'Startup mode', Integer(DefaultOptions.StartupMode)));

      VeriFingerMode := RegReadInteger(Reg, 'VeriFinger mode', DefaultOptions.VeriFingerMode);

      UseLowQualityImages := RegReadBool(Reg, 'Use low quality images', DefaultOptions.UseLowQualityImages);

      FileResolution := RegReadInteger(Reg, 'File resolution', DefaultOptions.FileResolution);
      FileFingerID := RegReadString(Reg, 'File finger ID', DefaultOptions.FileFingerID);
      FilePrompt := RegReadBool(Reg, 'File info prompt', DefaultOptions.FilePrompt);

      ScannerFingerID := RegReadString(Reg, 'Scanner finger ID', DefaultOptions.ScannerFingerID);
      ScannerPrompt := RegReadBool(Reg, 'Scanner info prompt', DefaultOptions.ScannerPrompt);

      CheckMinMinutiaCount := RegReadBool(Reg, 'Check min minutia count', DefaultOptions.CheckMinMinutiaCount);
      MinMinutiaCount := RegReadInteger(Reg, 'Min minutia count', DefaultOptions.MinMinutiaCount);
      SearchForDuplicates := RegReadBool(Reg, 'Search for duplicates', DefaultOptions.SearchForDuplicates);
      WarnDifferentFingerID := RegReadBool(Reg, 'Warn different finger id', DefaultOptions.WarnDifferentFingerID);

      UseG := RegReadBool(Reg, 'Use G', DefaultOptions.UseG);
      DatabasePercent := RegReadInteger(Reg, 'Database percent', DefaultOptions.DatabasePercent);
      MatchAll := RegReadBool(Reg, 'Match all database', DefaultOptions.MatchAll);
      ShowAllResults := RegReadBool(Reg, 'Show all results', DefaultOptions.ShowAllResults);

      ReturnedImage := RegReadInteger(Reg, 'Returned image', DefaultOptions.ReturnedImage);

      FARate := RegReadFloat(Reg, 'FA rate', DefaultOptions.FARate);
      MatchingSpeed := RegReadInteger(Reg, 'Matching speed', DefaultOptions.MatchingSpeed);
      MaximalRotation := RegReadInteger(Reg, 'Maximal rotation', DefaultOptions.MaximalRotation);

      GenFARate := RegReadFloat(Reg, 'Gen FA rate', DefaultOptions.GenFARate);
      GenMaximalRotation := RegReadInteger(Reg, 'Maximal rotation', DefaultOptions.MaximalRotation);
    end;
  finally
    Reg.Free;
  end;
end;

end.

